import debug from 'debug';
import { lensPath, set, splitEvery, update } from 'ramda';
import { FC, useEffect, useMemo, useState } from 'react';
import tw, { css, styled } from 'twin.macro';
import { Borders, borders } from '../../constants/border';
import { useLedCount } from '../contents/led-count';
import { DisplayConfig, LedStripConfigOfBorders } from '../models/display-config';
import { LedStripConfig } from '../models/led-strip-config';
import { PixelRgb } from '../models/pixel-rgb';
import { ScreenshotDto } from '../models/screenshot.dto';
import { CompletedContainer } from './completed-container';
import { DraggableStrip } from './draggable-strip';

export const logger = debug('app:completed-led-strip');

interface CompletedLedStripProps {
  screenshots: ScreenshotDto[];
  onDisplayConfigChange?: (value: DisplayConfig) => void;
}

export type BorderLedStrip = {
  pixels: PixelRgb[];
  config: LedStripConfig | null;
};

const StyledContainer = styled.section(
  ({ rows, columns }: { rows: number; columns: number }) => [
    tw`grid m-4 pb-2 items-center justify-items-center select-none`,
    tw`overflow-x-auto overflow-y-hidden flex-none`,
    css`
      grid-template-columns: repeat(${columns}, 0.5em);
      grid-template-rows: auto repeat(${rows}, 0.5em);
    `,
  ],
);
export const CompletedLedStrip: FC<CompletedLedStripProps> = ({
  screenshots,
  onDisplayConfigChange,
}) => {
  const borderLedStrips: BorderLedStrip[] = useMemo(() => {
    return screenshots.flatMap((ss) =>
      borders.map((b) => ({
        pixels: splitEvery(3, Array.from(ss.colors[b])) as PixelRgb[],
        config: ss.config.led_strip_of_borders[b],
      })),
    );
  }, [screenshots]);
  const ledCount = useMemo(
    () => borderLedStrips.reduce((prev, curr) => prev + curr.pixels.length, 0),
    [borderLedStrips],
  );
  const maxIndex = useMemo(
    () =>
      Math.max(
        ...borderLedStrips
          .map((s) => [
            s.config?.global_end_position ?? 0,
            s.config?.global_start_position ?? 0,
          ])
          .flat(),
      ),
    [borderLedStrips],
  );

  const { setLedCount } = useLedCount();
  // setLedCount for context
  useEffect(() => {
    setLedCount(ledCount);
  }, [ledCount, setLedCount]);

  const [overrideBorderLedStrips, setOverrideBorderLedStrips] =
    useState<BorderLedStrip[]>();

  const strips = useMemo(() => {
    return borderLedStrips.map(({ config, pixels }, index) =>
      config ? (
        <DraggableStrip
          key={index}
          {...{ config, pixels, index: index + 1 }}
          onConfigChange={(c) => {
            setOverrideBorderLedStrips(
              update(index, { config: c, pixels }, borderLedStrips),
            );
          }}
          onConfigFinish={(c) => {
            const indexOfDisplay = Math.floor(index / borders.length);
            const xLens = lensPath<LedStripConfigOfBorders, Borders>([
              borders[index % borders.length],
            ]);
            const displayConfig: DisplayConfig = {
              ...screenshots[indexOfDisplay].config,
              led_strip_of_borders: set(
                xLens,
                c,
                screenshots[indexOfDisplay].config.led_strip_of_borders,
              ),
            };
            onDisplayConfigChange?.(displayConfig);
          }}
        />
      ) : (
        <div key={index} />
      ),
    );
  }, [borderLedStrips, screenshots]);

  useEffect(() => {
    setOverrideBorderLedStrips(undefined);
  }, [borderLedStrips]);
  return (
    <StyledContainer rows={screenshots.length * borders.length} columns={maxIndex + 1}>
      <CompletedContainer
        borderLedStrips={borderLedStrips}
        overrideBorderLedStrips={overrideBorderLedStrips}
      />
      {strips}
    </StyledContainer>
  );
};
